#!/usr/bin/python
# -*- coding: utf-8 -*-

#
# PANDA3D ENGINE IMPORTS
#
from direct.interval.LerpInterval import LerpFunc
from direct.interval.IntervalGlobal import Sequence, Parallel, Func, Wait

__author__ = "Fireclaw the Fox"
__license__ = """
Simplified BSD (BSD 2-Clause) License.
See License.txt or http://opensource.org/licenses/BSD-2-Clause for more info
"""


#
# ANIMATION FSM HANDLER
#
class Animator:
    """
    This class is responsible for playing animations and easing between
    them. It also throws the events usable for playing some audio.
    """
    def __init__(self):
        self.current_animations = []
        self.enterSprintDuration = 1.5
        self.enterRunDuration = 1.5
        self.enterWalkDuration = 0.5
        self.pre_jump_state = self.STATE_IDLE
        self.pre_pause_anim = self.IDLE
        self.pre_pause_frame = 0
        self.current_seq = None
        self.wall_run_direction = self.WALLRUN_UP

        self.skipPlayRateChanges = [self.IDLE]

        def createEaseIn(anim):
            return LerpFunc(
                self.easeAnimation,
                fromData=0,
                toData=1,
                blendType="easeIn",
                extraArgs=[anim])

        def createEaseOut(anim):
            return LerpFunc(
                self.easeAnimation,
                fromData=1,
                toData=0,
                blendType="easeOut",
                extraArgs=[anim])

        self.ease_in_idle = createEaseIn(self.IDLE)
        self.ease_out_idle = createEaseOut(self.IDLE)

        self.ease_in_walk = createEaseIn(self.WALK)
        self.ease_out_walk = createEaseOut(self.WALK)

        self.ease_in_run = createEaseIn(self.RUN)
        self.ease_out_run = createEaseOut(self.RUN)

        self.ease_in_sprint = createEaseIn(self.SPRINT)
        self.ease_out_sprint = createEaseOut(self.SPRINT)

        self.ease_out_jump = createEaseOut(self.JUMP_START)
        self.ease_out_fall = createEaseOut(self.FALL)
        self.ease_out_land = createEaseOut(self.JUMP_LAND)

        self.ease_in_wall_run_up = createEaseIn(self.WALLRUN_UP)
        self.ease_in_wall_run_l = createEaseIn(self.WALLRUN_LEFT)
        self.ease_in_wall_run_r = createEaseIn(self.WALLRUN_RIGHT)


    def pauseAnimator(self):
        self.pre_pause_anim = self.getCurrentAnim()
        self.pre_pause_frame = self.getCurrentFrame(self.pre_pause_anim)
        if self.current_seq:
            self.current_seq.pause()
        # stop the current running animation
        self.stop()

    def resumeAnimator(self):
        if self.pre_pause_frame is None: self.pre_pause_frame = 0
        if self.pre_pause_anim is None: self.pre_pause_anim = self.IDLE
        self.pose(self.pre_pause_anim, self.pre_pause_frame)
        self.loop(self.pre_pause_anim, restart = False)
        if self.current_seq:
            if self.current_seq.getState() != Sequence.SFinal:
                self.current_seq.resume()

    def easeAnimation(self, t, anim):
        self.setControlEffect(anim, t)

    def setCurrentAnimsPlayRate(self, rate):
        """Set the play rate of the current playing animations to rate"""
        for anim in self.current_animations:
            if anim in self.skipPlayRateChanges:
                continue
            self.setPlayRate(rate, anim)
            base.messenger.send(self.audio_set_walk_playrate_evt, [rate])

    def tryRequest(self, state):
        if self.state != self.STATE_JUMP and not self.isInTransition():
            if not self.state == state:
                self.request(state)

    def startCurSeq(self, animFrom, animTo, easeIn, easeOut, finalState):
        if self.current_seq != None:
            self.current_seq.finish()
        self.current_seq = Sequence(
            Func(self.enableBlend),
            Func(self.loop, animTo),
            Parallel(
                easeOut,
                easeIn),
            Func(self.stop, animFrom),
            Func(self.disableBlend),
            Func(self.tryRequest, finalState))
        self.current_seq.start()

    def enterIdle(self):
        base.messenger.send(self.audio_stop_walk_evt)
        self.current_animations = [self.IDLE]
        if not self.getCurrentAnim() == self.IDLE:
            self.loop(self.IDLE)

    def enterWalk(self):
        self.current_animations = [self.WALK]
        if not self.getCurrentAnim() == self.WALK:
            self.loop(self.WALK)

    def enterRun(self):
        self.current_animations = [self.RUN]
        if not self.getCurrentAnim() == self.RUN:
            self.loop(self.RUN)

    def enterSprint(self):
        self.current_animations = [self.SPRINT]
        if not self.getCurrentAnim() == self.SPRINT:
            self.loop(self.SPRINT)

    def enterIdleToWalk(self):
        base.messenger.send(self.audio_play_walk_evt)
        self.current_animations = [self.IDLE, self.WALK]
        self.ease_out_idle.duration = self.enterWalkDuration
        self.ease_in_walk.duration = self.enterWalkDuration
        self.startCurSeq(self.IDLE, self.WALK, self.ease_in_walk, self.ease_out_idle, self.STATE_WALK)
    def exitIdleToWalk(self):
        self.stop(self.IDLE)
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterIdleToRun(self):
        base.messenger.send(self.audio_play_run_evt)
        self.current_animations = [self.IDLE, self.RUN]
        self.ease_out_idle.duration = self.enterRunDuration
        self.ease_in_run.duration = self.enterRunDuration
        self.startCurSeq(self.IDLE, self.RUN, self.ease_in_run, self.ease_out_idle, self.STATE_RUN)
    def exitIdleToRun(self):
        self.stop(self.IDLE)
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterIdleToSprint(self):
        base.messenger.send(self.audio_play_sprint_evt)
        self.current_animations = [self.IDLE, self.SPRINT]
        self.ease_out_idle.duration = self.enterSprintDuration
        self.ease_in_sprint.duration = self.enterSprintDuration
        self.startCurSeq(self.IDLE, self.SPRINT, self.ease_in_sprint, self.ease_out_idle, self.STATE_SPRINT)
    def exitIdleToSprint(self):
        self.stop(self.IDLE)
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterWalkToIdle(self):
        base.messenger.send(self.audio_play_walk_evt)
        self.current_animations = [self.WALK, self.IDLE]
        self.ease_out_walk.duration = self.current_accleration/self.current_max_accleration
        self.ease_in_idle.duration = self.current_accleration/self.current_max_accleration
        self.startCurSeq(self.WALK, self.IDLE, self.ease_in_idle, self.ease_out_walk, self.STATE_IDLE)
    def exitWalkToIdle(self):
        self.stop(self.WALK)
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterWalkToRun(self):
        base.messenger.send(self.audio_play_run_evt)
        self.current_animations = [self.WALK, self.RUN]
        self.startCurSeq(self.WALK, self.RUN, self.ease_in_run, self.ease_out_walk, self.STATE_RUN)
    def exitWalkToRun(self):
        self.stop(self.WALK)
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterRunToIdle(self):
        self.current_animations = [self.RUN, self.IDLE]
        self.ease_out_run.duration = self.current_accleration/self.current_max_accleration
        self.ease_in_idle.duration = self.current_accleration/self.current_max_accleration
        self.startCurSeq(self.RUN, self.IDLE, self.ease_in_idle, self.ease_out_run, self.STATE_IDLE)
    def exitRunToIdle(self):
        self.stop(self.RUN)
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterRunToWalk(self):
        base.messenger.send(self.audio_play_walk_evt)
        self.current_animations = [self.RUN, self.WALK]
        self.startCurSeq(self.RUN, self.WALK, self.ease_in_walk, self.ease_out_run, self.STATE_WALK)
    def exitRunToWalk(self):
        self.stop(self.RUN)
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterRunToSprint(self):
        base.messenger.send(self.audio_play_sprint_evt)
        self.current_animations = [self.RUN, self.SPRINT]
        self.startCurSeq(self.RUN, self.SPRINT, self.ease_in_sprint, self.ease_out_run, self.STATE_SPRINT)
    def exitRunToSprint(self):
        self.stop(self.RUN)
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterSprintToIdle(self):
        self.current_animations = [self.SPRINT, self.IDLE]
        self.ease_out_sprint.duration = self.current_accleration/self.current_max_accleration
        self.ease_in_idle.duration = self.current_accleration/self.current_max_accleration
        self.startCurSeq(self.SPRINT, self.IDLE, self.ease_in_idle, self.ease_out_sprint, self.STATE_IDLE)
    def exitSprintToIdle(self):
        self.stop(self.SPRINT)
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterSprintToRun(self):
        base.messenger.send(self.audio_play_run_evt)
        self.current_animations = [self.WALK, self.RUN]
        self.startCurSeq(self.SPRINT, self.RUN, self.ease_in_run, self.ease_out_sprint, self.STATE_RUN)
    def exitSprintToRun(self):
        self.stop(self.SPRINT)
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterJump(self):
        base.messenger.send(self.audio_stop_walk_evt)
        base.messenger.send(self.audio_play_jump_evt)
        self.current_animations = [self.JUMP_START]
        if not self.getCurrentAnim() == self.JUMP_START:
            self.current_seq = Sequence(
                self.actorInterval(self.JUMP_START, False),
                Func(self.tryRequest, self.STATE_FALL))
            self.current_seq.start()
    def exitJump(self):
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterFall(self):
        base.messenger.send(self.audio_stop_walk_evt)
        base.messenger.send(self.audio_play_fall_evt)
        if self.getCurrentAnim() != self.FALL and self.getCurrentAnim() != self.JUMP_START:
            self.loop(self.FALL)

    def enterLand(self):
        base.messenger.send(self.audio_play_land_evt)
        self.current_animations = [self.JUMP_LAND]
        next_state = self.STATE_IDLE
        if self.pre_jump_state == self.STATE_RUN \
        or self.pre_jump_state == self.STATE_IDLE_TO_RUN \
        or self.pre_jump_state == self.STATE_WALK_TO_RUN:
            self.LandToRun()
        elif self.pre_jump_state == self.STATE_WALK \
        or self.pre_jump_state == self.STATE_IDLE_TO_WALK \
        or self.pre_jump_state == self.STATE_RUN_TO_WALK:
            self.LandToWalk()
        else:
            self.current_seq = Sequence(
                self.actorInterval(self.JUMP_LAND, False),
                Func(self.tryRequest, self.STATE_IDLE))
            self.current_seq.start()
    def exitLand(self):
        self.ease_in_walk.duration = self.enterWalkDuration
        self.ease_in_run.duration = self.enterRunDuration
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def LandToWalk(self):
        base.messenger.send(self.audio_play_walk_evt)
        self.current_animations = [self.JUMP_LAND, self.WALK]
        self.ease_out_land.duration = 0.5
        self.ease_in_walk.duration = 0.5
        self.startCurSeq(self.JUMP_LAND, self.WALK, self.ease_in_walk, self.ease_out_land, self.STATE_WALK)

    def LandToRun(self):
        base.messenger.send(self.audio_play_run_evt)
        self.current_animations = [self.JUMP_LAND, self.RUN]
        self.ease_out_land.duration = 0.25
        self.ease_in_run.duration = 0.25
        self.startCurSeq(self.JUMP_LAND, self.RUN, self.ease_in_run, self.ease_out_land, self.STATE_RUN)

    def LandToSprint(self):
        base.messenger.send(self.audio_play_sprint_evt)
        self.current_animations = [self.JUMP_LAND, self.SPRINT]
        self.ease_out_land.duration = 0.25
        self.ease_in_sprint.duration = 0.25
        self.startCurSeq(self.JUMP_LAND, self.SPRINT, self.ease_in_sprint, self.ease_out_land, self.STATE_SPRINT)

    def setWallRunDirection(self, direction):
        self.wall_run_direction = direction

    def enterWallRun(self):
        self.current_animations = [self.wall_run_direction]
        if not self.getCurrentAnim() == self.wall_run_direction:
            self.loop(self.wall_run_direction)

    def startWRSeq(self, animFrom, easeOut):
        if self.wall_run_direction == self.WALLRUN_UP:
            self.startCurSeq(animFrom, self.WALLRUN_UP, self.ease_in_wall_run_up, easeOut, self.STATE_WALL_RUN)
        elif self.wall_run_direction == self.WALLRUN_LEFT:
            self.startCurSeq(animFrom, self.WALLRUN_LEFT, self.ease_in_wall_run_l, easeOut, self.STATE_WALL_RUN)
        elif self.wall_run_direction == self.WALLRUN_RIGHT:
            self.startCurSeq(animFrom, self.WALLRUN_RIGHT, self.ease_in_wall_run_r, easeOut, self.STATE_WALL_RUN)

    def enterRunToWallRun(self):
        base.messenger.send(self.audio_play_run_evt)
        self.startWRSeq(self.RUN, self.ease_out_run)
    def exitRunToWallRun(self):
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterSprintToWallRun(self):
        base.messenger.send(self.audio_play_run_evt)
        self.startWRSeq(self.SPRINT, self.ease_out_sprint)
    def exitSprintToWallRun(self):
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterJumpToWallRun(self):
        base.messenger.send(self.audio_play_run_evt)
        self.startWRSeq(self.JUMP_START, self.ease_out_jump)
    def exitJumpToWallRun(self):
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterFallToWallRun(self):
        base.messenger.send(self.audio_play_run_evt)
        self.startWRSeq(self.FALL, self.ease_out_fall)
    def exitFallToWallRun(self):
        if self.current_seq is not None:
            self.current_seq.finish()
            self.current_seq = None

    def enterLedgeGrab(self):
        self.loop(self.LEDGE_GRAB)
        
    def enterLedgeGrabUp(self):
        self.current_seq = Sequence(
            self.actorInterval(self.LEDGE_GRAB_UP),
            Func(self.tryRequest, self.STATE_IDLE))
        self.current_seq.start()
        
    def enterLedgeGrabLeft(self):
        self.current_animations = [self.LEDGE_GRAB_LEFT]
        if not self.getCurrentAnim() == self.LEDGE_GRAB_LEFT:
            self.loop(self.LEDGE_GRAB_LEFT)

    def enterLedgeGrabRight(self):
        self.current_animations = [self.LEDGE_GRAB_RIGHT]
        if not self.getCurrentAnim() == self.LEDGE_GRAB_RIGHT:
            self.loop(self.LEDGE_GRAB_RIGHT)
