#!/usr/bin/python
# -*- coding: utf-8 -*-

#
# PANDA3D ENGINE IMPORTS
#
from panda3d.core import PandaNode

__author__ = "Fireclaw the Fox"
__license__ = """
Simplified BSD (BSD 2-Clause) License.
See License.txt or http://opensource.org/licenses/BSD-2-Clause for more info
"""


#
# CAMERA FUNCTIONS
#
class CameraThirdPerson:
    def __init__(self, parent, near=0.5, far=5000, fov=90):
        self.parent = parent
        # set the minimum and maximum view distance
        base.camLens.setNearFar(near, far)
        base.camLens.setFov(fov)

        self.last_platform_position_cam = None
        self.last_platform_rotation_cam = None

    def startCamera(self):
        """Starts the camera module."""
        # to enhance readability
        self.cam_floater = self.parent.mainNode
        self.cam_floater = self.parent.mainNode.attachNewNode(PandaNode("playerCamFloater"))
        self.cam_floater.setPos(self.parent.cam_floater_pos)
        taskMgr.add(self.updateCamera, "task_camActualisation", priority=-4)

    def stopCamera(self):
        taskMgr.remove("task_camActualisation")
        self.cam_floater.removeNode()

    def pauseCamera(self):
        taskMgr.remove("task_camActualisation")

    def resumeCamera(self):
        taskMgr.add(self.updateCamera, "task_camActualisation", priority=-4)

    def centerCamera(self):
        """This method will move the camera centered behind the player model"""
        # Camera Movement Updates
        camvec = self.cam_floater.getPos() - camera.getPos()
        camvec.setZ(0)
        camdist = camvec.length()
        camdist = 3.0
        # get the cameras current offset to the player model on the z-axis
        offsetZ = self.cam_floater.getZ()
        camera.setPos(self.parent.mainNode, 0, camdist, offsetZ)
        base.win.movePointer(0, self.parent.win_width_half, self.parent.win_height_half)
        camera.lookAt(self.cam_floater)

    def updateCamera(self, task):
        """This function will check the min and max distance of the camera to
        the defined model and will correct the position if the cam is to close
        or to far away"""

        # Camera Movement Updates
        camvec = self.cam_floater.getPos(render) - camera.getPos(render)
        camvec.setZ(0)
        camdist = camvec.length()
        camvec.normalize()

        # check if the camera should be centered
        if self.parent.isDown(self.parent.key_center_camera):
            if base.transitions.letterboxIval:
                if not base.transitions.letterboxIval.isPlaying():
                    base.transitions.letterboxOn(0.15)
            else:
                base.transitions.letterboxOn(0.15)
            self.centerCamera()
            return task.cont

        if base.transitions.letterbox is not None:
            if not base.transitions.letterboxIval.isPlaying() \
            and not base.transitions.letterbox.isStashed():
                base.transitions.letterboxOff(0.15)

        dt = globalClock.getDt()

        # Move camera left/right with the mouse
        if base.mouseWatcherNode.hasMouse() and self.parent.enable_mouse:
            mw = base.mouseWatcherNode
            if base.win.movePointer(0, self.parent.win_width_half, self.parent.win_height_half):
                # Vertical positioning
                mouse_y = mw.getMouseY()
                z = mouse_y * self.parent.mouse_speed_y * camdist * dt
                if self.parent.mouse_invert_vertical:
                    camera.setZ(camera, z)
                else:
                    camera.setZ(camera, -z)

                # Horizontal positioning
                mouse_x = mw.getMouseX()
                x = mouse_x * self.parent.mouse_speed_x * camdist * dt
                if self.parent.mouse_invert_horizontal:
                    camera.setX(camera, x)
                else:
                    camera.setX(camera, -x)

        # Move camera with the keyboard
        if self.parent.isDown(self.parent.key_camera_left):
            x = self.parent.keyboard_cam_speed_x * dt
            if self.parent.keyboard_invert_horizontal:
                x = -x
            camera.setX(camera, -x)
        if self.parent.isDown(self.parent.key_camera_right):
            x = self.parent.keyboard_cam_speed_x * dt
            if self.parent.keyboard_invert_horizontal:
                x = -x
            camera.setX(camera, x)
        if self.parent.isDown(self.parent.key_camera_up):
            z = self.parent.keyboard_cam_speed_y * dt
            if self.parent.keyboard_invert_vertical:
                z = -z
            camera.setZ(camera, z)
        elif self.parent.isDown(self.parent.key_camera_down):
            z = self.parent.keyboard_cam_speed_y * dt
            if self.parent.keyboard_invert_vertical:
                z = -z
            camera.setZ(camera, -z)

        # Move camera with moving platforms
        if self.parent.active_platform is not None:
            platformPositionAbsolute = self.parent.active_platform.getPos()
            # Character on moving platform
            if self.last_platform_position_cam is None:
                self.last_platform_position_cam = platformPositionAbsolute
            platform_speed = platformPositionAbsolute - self.last_platform_position_cam
            self.last_platform_position_cam = platformPositionAbsolute
            # now update the player position according to the platform
            camera.setPos(camera.getPos()+platform_speed)
        else:
            self.last_platform_position_cam = None
            self.last_platform_rotation_cam = None

        # Get the cameras current offset to the player model on the z-axis
        offsetZ = camera.getZ(render) - self.cam_floater.getZ(render)
        # check if the camera is within the min and max z-axis offset
        if offsetZ < self.parent.min_cam_height_distance:
            # the cam is to low, so move it up
            camera.setZ(self.cam_floater.getZ(render) + self.parent.min_cam_height_distance)
            offsetZ = self.parent.min_cam_height_distance
        elif offsetZ > self.parent.max_cam_height_distance:
            # the cam is to high, so move it down
            camera.setZ(self.cam_floater.getZ(render) + self.parent.max_cam_height_distance)
            offsetZ = self.parent.max_cam_height_distance

        # lazy camera positioning
        # if we are not moving up or down, set the cam to an average position
        if offsetZ > self.parent.cam_height_avg_up:
            # the cam is higher then the average cam height above the player
            # so move it slowly down
            camera.setZ(camera.getZ(render) - self.parent.cam_z_justification_speed * globalClock.getDt())
            newOffsetZ = camera.getZ(render) - self.cam_floater.getZ()
            # check if the cam has reached the desired offset
            #if newOffsetZ < self.parent.cam_height_avg_down:
            #    # set the cam z position to exactly the desired offset
            #    camera.setZ(self.cam_floater.getZ(render) + self.parent.cam_height_avg_down)
        elif offsetZ < self.parent.cam_height_avg_down:
            # the cam is lower then the average cam height above the player
            # so move it slowly up
            camera.setZ(camera.getZ() + self.parent.cam_z_justification_speed * globalClock.getDt())
            newOffsetZ = camera.getZ() - self.cam_floater.getZ(render)
            # check if the cam has reached the desired offset
            #if newOffsetZ > self.parent.cam_height_avg_up:
            #    # set the cam z position to exactly the desired offset
            #    camera.setZ(self.cam_floater.getZ(render) + self.parent.cam_height_avg_up)

        # If the camera is to far from player start following
        if camdist > self.parent.max_cam_distance:
            camera.setPos(camera.getPos()+camvec*(camdist-self.parent.max_cam_distance))
            camdist = self.parent.max_cam_distance

        # camera collision detection
        # always set the cameras position to the first hitpoint on a collision solid
        had_ray_collision = False
        entry = self.parent.getFirstCollisionEntryInLine(self.cam_floater.getPos(render), camera.getPos(render))
        if entry is not None:
            pos = None
            if entry.hasSurfacePoint():
                pos = entry.getSurfacePoint(render)
            wall_normal = None
            if entry.hasSurfaceNormal():
                wall_normal = entry.getSurfaceNormal(render)
                pos.setX(pos.getX() + wall_normal.getX()/2.0)
                pos.setY(pos.getY() + wall_normal.getY()/2.0)
            #pos = self.parent.getFirstCollisionInLine(self.cam_floater.getPos(render), camera.getPos(render))
            if pos is not None:
                had_ray_collision = True
                camera.setPos(pos)

        # If player is to close move the camera backwards
        if camdist < self.parent.min_cam_distance:
            if not had_ray_collision:
                # move the camera backwards
                camera.setPos(camera.getPos()-camvec*(self.parent.min_cam_distance-camdist))
                camdist = self.parent.min_cam_distance
            else:
                # we can't move the camera back into the wall, so move it up
                #TODO: Maybe change max_cam_height_distance with something like
                #      self.parent.min_cam_dist - camdist
                camera.setZ(self.cam_floater.getZ() + self.parent.max_cam_height_distance)

        camera.lookAt(self.cam_floater)
        return task.cont
