#!/usr/bin/python
# -*- coding: utf-8 -*-

#
# PYTHON IMPORTS
#
import math

#
# PANDA3D ENGINE IMPORTS
#
from panda3d.core import Vec3, Point3F, Point3

__author__ = "Fireclaw the Fox"
__license__ = """
Simplified BSD (BSD 2-Clause) License.
See License.txt or http://opensource.org/licenses/BSD-2-Clause for more info
"""


#
# MOVE FUNCTIONS
#
class Control:
    """This is the main mover class.

    It is responsible for handling player input and acting according to
    the input.
    This class handles anything that isn't physics and collision related
    but glues together most of the other classes and is responsible for
    requesting most of the animation states and calling the physics
    update methods"""

    def __init__(self):
        # the actual accleration the character is at
        self.current_accleration = 0.0
        self.current_max_accleration = 0.0
        self.wasJumping = False
        self.fall_time = 0.0
        self.last_platform_position = None
        self.last_platform_rotation = None
        self.movementVec = Vec3(0, 0, 0)

    def startControl(self):
        """Start the control module"""
        taskMgr.add(self.move, "task_movement", priority=-15)

    def stopControl(self):
        """Stops the controller class"""
        taskMgr.remove("task_movement")

    def pauseControl(self):
        """Stops the controls from being usable by the gamer"""
        self.stopControl()
        if self.state in self.on_ground_states:
            self.current_accleration = 0.0
            self.current_max_accleration = 0.0

    def move(self, task):
        """The main task for updating the players position according
        to the keys pressed by the user"""
        if self.state in self.ignore_input_states:
            return task.cont


        dt = globalClock.getDt()
        isMoving = False
        request_state = None
        accleration_mult = 1.0
        isMoving = self.current_accleration > 0
        self.current_max_accleration = self.max_accleration_run
        rotation = None
        isAirborn = False

        if self.state not in self.on_ground_states \
        and self.state not in self.flying_states:
            # We are not on the ground and hence can't move
            isAirborn = True
            self.fall_time += dt
            if self.fall_time > 1.0 and self.state != self.STATE_FALL:
                request_state = self.STATE_FALL
        elif self.wasJumping:
            self.wasJumping = False
            isAirborn = False
            self.fall_time = 0.0
            self.current_accleration = self.pre_jump_accleration / 2.0
        else:
            isAirborn = False
            self.fall_time = 0.0

        #
        # PLAYER MOVEMENT
        #
        y_vec = -1 if self.first_pserson_mode else 1
        if self.isDown(self.key_forward):
            if self.movementVec.getY() == y_vec:
                self.current_accleration /= 4.0
            self.movementVec.setY(-y_vec)
        elif self.isDown(self.key_backward):
            if self.movementVec.getY() == -y_vec:
                self.current_accleration /= 4.0
            self.movementVec.setY(y_vec)
        else:
            self.movementVec.setY(0)

        if self.isDown(self.key_left):
            if self.movementVec.getX() == 1:
                self.current_accleration /= 4.0
            self.movementVec.setX(-1)
        elif self.isDown(self.key_right):
            if self.movementVec.getX() == -1:
                self.current_accleration /= 4.0
            self.movementVec.setX(1)
        else:
            self.movementVec.setX(0)

        moveKeyPressed = self.movementVec != Vec3()
        if isAirborn:
            if self.current_accleration > 0:
                self.current_accleration -= (self.deaccleration / 4.0) * dt
                if self.current_accleration < 0:
                    self.current_accleration = 0
                # calculate the players heading
            if not self.isDown(self.key_center_camera) \
            and moveKeyPressed \
            and not self.first_pserson_mode:
                angle = math.atan2(self.movementVec.getX(), self.movementVec.getY())
                rotation = angle * (180.0 / math.pi)
        elif moveKeyPressed:
            isMoving = True
            if not self.isDown(self.key_center_camera) \
            and not self.first_pserson_mode:
                # calculate the players heading
                angle = math.atan2(self.movementVec.getX(), self.movementVec.getY())
                rotation = angle * (180.0 / math.pi)

            # set the players accleration
            if self.isDown(self.key_sprint):
                # sprint
                accleration = self.accleration_sprint
                self.current_max_accleration = self.max_accleration_sprint
            elif self.isDown(self.key_walk):
                # walk
                # set accleration
                accleration = self.accleration_walk
                self.current_max_accleration = self.max_accleration_walk
            else:
                # run
                accleration = self.accleration_run
            # acclerate until we reached the maximum speed
            if self.current_accleration < self.current_max_accleration:
                self.current_accleration += accleration * dt
                if self.current_accleration > self.current_max_accleration:
                    self.current_accleration = self.current_max_accleration
            elif self.current_accleration > self.current_max_accleration:
                self.current_accleration -= self.deaccleration * dt
                if self.current_accleration < self.current_max_accleration:
                    self.current_accleration = self.current_max_accleration
        elif self.current_accleration > 0:
            if self.state == self.STATE_WALK:
                request_state = self.STATE_WALK_TO_IDLE
            elif self.state == self.STATE_RUN:
                request_state = self.STATE_RUN_TO_IDLE
            elif self.state == self.STATE_SPRINT:
                request_state = self.STATE_SPRINT_TO_IDLE
            # deacclerate until we stopped completely again
            self.current_accleration -= self.deaccleration * dt
            if self.current_accleration < 0:
                self.current_accleration = 0

        #
        # PLAYER STATES
        #
        if isAirborn:
            pass
        elif self.isDown(self.key_sprint) and moveKeyPressed and isMoving:
            # set animation
            self.current_max_accleration = self.max_accleration_run
            if self.state == self.STATE_IDLE:
                request_state = self.STATE_IDLE_TO_SPRINT
            elif self.state == self.STATE_RUN:
                request_state = self.STATE_RUN_TO_SPRINT
            elif self.state == self.STATE_RUN_TO_IDLE:
                request_state = self.STATE_RUN_TO_SPRINT
            elif self.state == self.STATE_SPRINT_TO_IDLE:
                request_state = self.STATE_SPRINT
            elif self.state == self.STATE_WALK_TO_IDLE or self.state == self.STATE_WALK:
                request_state = self.STATE_WALK_TO_RUN
        elif self.isDown(self.key_walk) and moveKeyPressed and isMoving:
            # set animation
            if self.state == self.STATE_IDLE:
                request_state = self.STATE_IDLE_TO_WALK
            elif self.state == self.STATE_RUN:
                request_state = self.STATE_RUN_TO_WALK
            elif self.state == self.STATE_WALK_TO_IDLE:
                request_state = self.STATE_WALK
            elif self.state == self.STATE_RUN_TO_IDLE:
                request_state = self.STATE_WALK
            elif self.state == self.STATE_SPRINT or self.state == self.STATE_SPRINT_TO_IDLE:
                request_state = self.STATE_SPRINT_TO_RUN
        elif isMoving:
            # no modifier key pressed, so simply run
            # set animation
            self.current_max_accleration = self.max_accleration_run
            if self.state == self.STATE_IDLE and moveKeyPressed:
                request_state = self.STATE_IDLE_TO_RUN
            elif self.state == self.STATE_WALK and moveKeyPressed:
                request_state = self.STATE_WALK_TO_RUN
            elif self.state == self.STATE_WALK_TO_IDLE and moveKeyPressed:
                request_state = self.STATE_RUN
            elif self.state == self.STATE_RUN_TO_IDLE and moveKeyPressed:
                request_state = self.STATE_RUN
            elif self.state == self.STATE_SPRINT and moveKeyPressed:
                request_state = self.STATE_SPRINT_TO_RUN

        if isAirborn:
            pass
        elif self.current_accleration <= 0:
            if self.state == self.STATE_WALK:
                request_state = self.STATE_WALK_TO_IDLE
            elif self.state == self.STATE_RUN:
                request_state = self.STATE_RUN_TO_IDLE
            elif self.state == self.STATE_SPRINT:
                request_state = self.STATE_SPRINT_TO_IDLE
            elif self.state != self.STATE_IDLE and self.state != self.STATE_LAND:
                request_state = self.STATE_IDLE
        else:
            self.setCurrentAnimsPlayRate(self.current_accleration/self.current_max_accleration)

        #
        # COLLISION CHECK PREPARATIONS
        #
        tmp_np = self.mainNode.attachNewNode("temporary")

        #
        # WALL COLLISION CHECKS WALL RUN
        #
        # Wall checks to the front, left and right side of the player
        tmp_np.setPos(self.mainNode, 0, -self.wall_run_forward_check_dist, 0)
        point_a = self.mainNode.getPos()
        point_a.setZ(self.mainNode.getZ() + self.player_height/2.0)
        point_b = tmp_np.getPos(render)
        point_b.setZ(self.mainNode.getZ() + self.player_height/2.0)
        char_front_collision = self.getFirstCollisionInLine(point_a, point_b)
        # Wall run forward collision
        tmp_np.setPos(self.mainNode, 0, -self.wall_run_forward_check_dist, 0)
        point_b = tmp_np.getPos(render)
        point_b.setZ(self.mainNode.getZ() + 0.1)
        char_front_collision_entry = self.getFirstCollisionEntryInLine(point_a, point_b)
        # Left side collision
        tmp_np.setPos(self.mainNode, self.wall_run_sideward_check_dist, 0, 0)
        point_b = tmp_np.getPos(render)
        point_b.setZ(self.mainNode.getZ() + 0.1)
        char_left_collision_entry = self.getFirstCollisionEntryInLine(point_a, point_b)
        # Right side collision
        tmp_np.setPos(self.mainNode, -self.wall_run_sideward_check_dist, 0, 0)
        point_b = tmp_np.getPos(render)
        point_b.setZ(self.mainNode.getZ() + 0.1)
        char_right_collision_entry = self.getFirstCollisionEntryInLine(point_a, point_b)
        intel_action = []
        #
        # WALL RUN LOGIC
        #
        if (
            char_front_collision_entry is not None \
            or char_left_collision_entry is not None \
            or char_right_collision_entry is not None \
        ) \
        and self.isDown(self.key_intel_action) \
        and self.state in self.wall_run_possible_states:
            wall_run_possible = False
            wall_normal = None
            if char_front_collision_entry:
                # we have a wall in front of us that we can walk up
                if char_front_collision_entry.hasSurfaceNormal():
                    wall_normal = char_front_collision_entry.getSurfaceNormal(render)
                    self.setWallRunDirection(self.WALLRUN_UP)
            elif char_left_collision_entry:
                # we have a wall to our left that we can walk along
                if char_left_collision_entry.hasSurfaceNormal():
                    wall_normal = char_left_collision_entry.getSurfaceNormal(render)
                    wall_normal.setY(-wall_normal.getY())
                    wall_normal.setX(-wall_normal.getX())
                    self.setWallRunDirection(self.WALLRUN_LEFT)
            elif char_right_collision_entry:
                # we have a wall to our right that we can walk along
                if char_right_collision_entry.hasSurfaceNormal():
                    wall_normal = char_right_collision_entry.getSurfaceNormal(render)
                    self.setWallRunDirection(self.WALLRUN_RIGHT)

            # set the characters heading if aplicable
            if wall_normal is not None:
                zx = math.atan2(wall_normal.getZ(), wall_normal.getX())*180/math.pi
                zy = math.atan2(wall_normal.getZ(), wall_normal.getY())*180/math.pi
                zx = abs(zx-90)
                zy = abs(zy-90)
                if zy >= self.min_wall_angle_for_wall_run and zx >= self.min_wall_angle_for_wall_run:
                    if self.wall_run_direction == self.WALLRUN_UP:
                        # face towards the wall
                        h = math.atan2(-wall_normal.getX(), wall_normal.getY())*180/math.pi
                    else:
                        # face along the wall
                        h = math.atan2(wall_normal.getY(), wall_normal.getX())*180/math.pi
                    self.updatePlayerHpr((h, 0, 0))
                    # invalidate other rotational movements that were set before
                    rotation = None
                    wall_run_possible = True
            if wall_run_possible:
                intel_action.append(self.INTEL_ACTION_WALL_RUN)
                self.pre_jump_state = self.STATE_RUN
                if self.state not in [self.STATE_WALL_RUN, self.STATE_RUN_TO_WALL_RUN, self.STATE_SPRINT_TO_WALL_RUN, self.STATE_JUMP]:
                    if self.state in [self.STATE_RUN, self.STATE_SPRINT_TO_RUN]:
                        request_state = self.STATE_RUN_TO_WALL_RUN
                    elif self.state in  [self.STATE_RUN_TO_SPRINT, self.STATE_SPRINT]:
                        request_state = self.STATE_SPRINT_TO_WALL_RUN
                    else:
                        request_state = self.STATE_WALL_RUN
                    self.jump_strength = self.wall_run_off_jump_strength
        else:
            self.jump_strength = self.jump_strength_default
            if self.state in self.wall_run_states:
                request_state = self.STATE_RUN

        #
        # LEDGE GRAB LOGIC
        #
        tmp_np.setPos(self.mainNode, 0, -self.ledge_forward_check_dist, self.ledge_top_check_dist)
        point_a = tmp_np.getPos(render)
        point_b = tmp_np.getPos(render)
        point_b.setZ(self.mainNode.getZ(render) + self.ledge_bottom_check_dist)
        ledge_collision = self.getFirstCollisionEntryInLine(point_a, point_b)

        # this variable can be used to check if the player can move left/right
        # when hanging on a ledge
        ledge_grab_can_move = False

        if self.state in self.ledge_grab_states \
        and (self.isDown(self.key_backward) \
        or self.isDown(self.key_jump) \
        or ledge_collision is None):
            rotation = None
            request_state = self.STATE_FALL
            isAirborn = True
            if self.INTEL_ACTION_WALL_RUN in intel_action:
                intel_action.remove(self.INTEL_ACTION_WALL_RUN)
            intel_action.append(self.INTEL_ACTION_LEDGE_GRAB)
            if self.isDown(self.key_backward):
                # let go from ledge grab
                self.toggleFlyMode(False)
            elif (self.isDown(self.key_jump) \
            or self.isDown(self.key_forward)) \
            and ledge_collision is not None:
                # we want to pull up
                pos = None
                if ledge_collision.hasSurfacePoint():
                    pos = ledge_collision.getSurfacePoint(render)
                elif ledge_collision.hasContactPos():
                    pos = ledge_collision.getContactPos(render)
                if pos is not None:
                    # we want to pull up on a ledge
                    request_state = self.STATE_LEDGE_GRAB_UP
                    self.updatePlayerPosFix(pos)
                    isAirborn = False
                else:
                    # we have no wall or place to pull up to, so simply
                    # let go from ledge and fall down
                    self.toggleFlyMode(False)
        elif self.state in self.ledge_grab_states \
        and (self.isDown(self.key_left) \
        or self.isDown(self.key_right)):
            if self.INTEL_ACTION_WALL_RUN in intel_action:
                intel_action.remove(self.INTEL_ACTION_WALL_RUN)
            intel_action.append(self.INTEL_ACTION_LEDGE_GRAB)
            rotation = None
            ledge_move_collision = None
            direction = None
            if self.isDown(self.key_left):
                tmp_np.setPos(self.mainNode, self.player_radius, -self.ledge_forward_check_dist, self.ledge_top_check_dist)
                point_a = tmp_np.getPos(render)
                point_b = tmp_np.getPos(render)
                point_b.setZ(self.mainNode.getZ(render) + self.ledge_bottom_check_dist)
                ledge_move_collision = self.getFirstCollisionEntryInLine(point_a, point_b)
                direction = "left"
            else:
                tmp_np.setPos(self.mainNode, -self.player_radius, -self.ledge_forward_check_dist, self.ledge_top_check_dist)
                point_a = tmp_np.getPos(render)
                point_b = tmp_np.getPos(render)
                point_b.setZ(self.mainNode.getZ(render) + self.ledge_bottom_check_dist)
                ledge_move_collision = self.getFirstCollisionEntryInLine(point_a, point_b)
                direction = "right"
            if ledge_move_collision is not None:
                ledge_grab_can_move = True
                self.setCurrentAnimsPlayRate(1.0)
                if direction == "left":
                    if self.state != self.STATE_LEDGE_GRAB_LEFT:
                        request_state = self.STATE_LEDGE_GRAB_LEFT
                elif direction == "right":
                    if self.state != self.STATE_LEDGE_GRAB_RIGHT:
                        request_state = self.STATE_LEDGE_GRAB_RIGHT
            elif self.state != self.STATE_LEDGE_GRAB:
                request_state = self.STATE_LEDGE_GRAB

            if char_front_collision_entry is not None \
            and char_front_collision_entry.hasSurfaceNormal():
                wall_normal = char_front_collision_entry.getSurfaceNormal(render)

                # set the characters heading
                zx = math.atan2(wall_normal.getZ(), wall_normal.getX())*180/math.pi
                zy = math.atan2(wall_normal.getZ(), wall_normal.getY())*180/math.pi
                zx = abs(zx-90)
                zy = abs(zy-90)
                # face towards the wall
                h = math.atan2(-wall_normal.getX(), wall_normal.getY())*180/math.pi
                self.updatePlayerHpr((h, 0, 0))
        elif self.state in self.ledge_grab_states \
        and ledge_collision is not None:
            # stay in ledge grab mode
            isAirborn = True
            rotation = None
            self.toggleFlyMode(True)
            if self.INTEL_ACTION_WALL_RUN in intel_action:
                intel_action.remove(self.INTEL_ACTION_WALL_RUN)
            intel_action.append(self.INTEL_ACTION_LEDGE_GRAB)
            if self.state != self.STATE_LEDGE_GRAB:
                request_state = self.STATE_LEDGE_GRAB
            else:
                request_state = None
        elif ledge_collision is not None \
        and self.isDown(self.key_intel_action) \
        and self.state != self.STATE_LEDGE_GRAB \
        and self.state in self.ledge_grab_possible_states:
            # initiate a ledge grab if certain conditions are met
            ledge_normal = None
            if ledge_collision.hasSurfaceNormal():
                ledge_normal = ledge_collision.getSurfaceNormal(render)
            if ledge_normal is not None:
                if ledge_normal.getZ() > 0:
                    isAirborn = True
                    rotation = None
                    self.toggleFlyMode(True)
                    if self.INTEL_ACTION_WALL_RUN in intel_action:
                        intel_action.remove(self.INTEL_ACTION_WALL_RUN)
                    intel_action.append(self.INTEL_ACTION_LEDGE_GRAB)
                    newPos = char_front_collision
                    z = Point3()
                    if ledge_collision.hasSurfacePoint():
                        z = ledge_collision.getSurfacePoint(render).getZ()
                    elif ledge_collision.hasContactPos():
                        z = ledge_collision.getContactPos(render).getZ()
                    if newPos is None:
                        newPos = ledge_collision.getSurfacePoint(render)
                    newPos.setZ(z - self.ledge_z_pos)

                    self.updatePlayerPosFix(newPos)
                    self.updatePlayerPosFix(self.player_radius+0.01, self.mainNode)
                    if self.state != self.STATE_LEDGE_GRAB:
                        request_state = self.STATE_LEDGE_GRAB

        #
        # FRONT WALL COLLISION AVOIDANCE
        #
        # Forward facing wall checks to not run head first into a wall
        # calculate the distance to a collision within the forward check distance
        tmp_np.setPos(self.mainNode, 0, -self.forward_check_distance, 0)
        point_a = self.mainNode.getPos()
        point_a.setZ(self.mainNode.getZ() + self.player_height/2.0)
        point_b = tmp_np.getPos(render)
        point_b.setZ(self.mainNode.getZ() + self.player_height/2.0)
        char_front_collision = self.getFirstCollisionInLine(point_a, point_b)
        # check this only if we are not interested in a wall trick
        if char_front_collision is not None \
        and self.INTEL_ACTION_WALL_RUN not in intel_action \
        and self.INTEL_ACTION_LEDGE_GRAB not in intel_action:
            if not isAirborn:
                colvec = char_front_collision - self.mainNode.getPos()
                colvec.setZ(0)
                coldist = colvec.length()
                doStateCheck = True
                if self.first_pserson_mode \
                and (self.isDown(self.key_backward)\
                or self.isDown(self.key_left) \
                or self.isDown(self.key_right)):
                  doStateCheck = False
                elif coldist <= self.forward_stop_distance:
                    # we have a wall in front of us, hence we can't move forward
                    self.current_accleration = 0
                elif moveKeyPressed:
                    self.current_accleration -= self.deaccleration * dt
                    if self.current_accleration <= 0:
                        self.current_accleration = self.forward_min_speed_to_stop
                        if self.state == self.STATE_IDLE:
                            request_state = self.STATE_IDLE_TO_WALK
                    doStateCheck = False
                if request_state != self.STATE_JUMP and doStateCheck:
                    if self.state == self.STATE_WALK:
                        request_state = self.STATE_WALK_TO_IDLE
                    elif self.state == self.STATE_RUN:
                        request_state = self.STATE_RUN_TO_IDLE
                    elif self.state == self.STATE_SPRINT:
                        request_state = self.STATE_SPRINT_TO_IDLE
                    else:
                        if self.STATE_IDLE in self.defaultTransitions[self.state]:
                            request_state = self.STATE_IDLE
                        else:
                            request_state = None
        else:
            self.movementVec.setZ(0)

        #
        # PLAYER POSITION UPDATE
        #
        # calculate the players forward movement speed
        platform_speed = Vec3(0,0,0)
        if self.active_platform is not None:
            platformPositionAbsolute = self.active_platform.getPos()
            # Character on moving platform
            if self.last_platform_position is None:
                self.last_platform_position = platformPositionAbsolute
            platform_speed = platformPositionAbsolute - self.last_platform_position
            self.last_platform_position = platformPositionAbsolute
            # now update the player position according to the platform
            if self.state in self.flying_states:
                self.updatePlayerPosFloating(platform_speed)
            else:
                self.updatePlayerPosFloatingFlyign(platform_speed)

            # check for the platforms rotation
            if self.respect_platform_rotation:
                if self.last_platform_rotation is None:
                    self.last_platform_rotation = self.active_platform.getH()
                platform_rotation = self.active_platform.getH() - self.last_platform_rotation
                self.last_platform_rotation = self.active_platform.getH()
                if self.state in self.flying_states:
                    self.updatePlayerHprFloating(platform_rotation, self.active_platform)
                else:
                    self.updatePlayerHprFloating(platform_rotation, self.active_platform)
        else:
            self.last_platform_position = None
            self.last_platform_rotation = None

        #
        # INITIATE JUMPING
        #
        if self.isDown(self.key_jump) \
        and self.jump_enabled \
        and (not isAirborn \
        or self.fall_time <= self.jump_allow_after_fall_time):
            if self.STATE_JUMP in self.defaultTransitions[self.state]:
                self.wasJumping = True
                isAirborn = True
                self.pre_jump_state = self.state
                #TODO: distinguish between wall jumps and normal jumps
                # forward wall run = jump up or back?
                # sidward wall run = jump in the oposite direction of the wall
                if self.platform_movement_affects_jump:
                    self.doJump(self.speed * self.current_accleration, platform_speed / dt)
                else:
                    self.doJump(self.speed * self.current_accleration)
                self.pre_jump_accleration = self.current_accleration
                self.current_accleration = 0.0
                request_state = self.STATE_JUMP

        if request_state:
            self.request(request_state)

        #
        # CALCULATE MOVEMENT SPEED
        #
        update_speed = Point3()
        if isAirborn and self.state not in self.flying_states:
            # character is jumping/falling/flying
            if moveKeyPressed:
                self.current_speed = max(
                    self.speed_airborn * dt,
                    self.speed * self.current_accleration * dt)
                update_speed = Point3(0, -self.current_speed, 0)
            else:
                self.current_speed = self.speed * self.current_accleration * dt
                update_speed = Point3(0, -self.current_speed, 0)
        elif self.state not in self.flying_states:
            # normal walking/running
            self.current_speed = self.speed * self.current_accleration * dt
            update_speed = Point3(0, -self.current_speed, 0)
        if self.first_pserson_mode:
            # as we don't rotate the character in first person mode, move
            # him in the direction of the movement vector
            update_speed = self.movementVec * update_speed.getY()

        #
        # CHECK FOR INTELIGENT ACTION MOVEMENT SPEEDS
        #
        if self.INTEL_ACTION_WALL_RUN in intel_action and not self.wasJumping:
            # if we are runing on a wall add some speed to the Z axis
            # movement of the player
            wr_speed = self.wall_run_speed * self.current_accleration
            if wr_speed > self.max_wall_run_speed:
                wr_speed = self.max_wall_run_speed
            #if wr_speed <= 0:
            #    wr_speed = self.min_wall_run_speed
            update_speed.setZ(wr_speed * dt)

        if self.INTEL_ACTION_LEDGE_GRAB in intel_action and ledge_grab_can_move:
            # in ledge grab mode we can only move left and right but not
            # rotate the player
            lg_speed = self.ledge_grab_sidward_move_speed
            if self.isDown(self.key_left):
                update_speed.setX(lg_speed * dt)
            elif self.isDown(self.key_right):
                update_speed.setX(-lg_speed * dt)

        #
        # COLLISION CHECK CLEANUP
        #
        tmp_np.removeNode()

        #
        # UPDATE IN THE PHYSICS CLASS
        #
        ignore_step = self.INTEL_ACTION_WALL_RUN in intel_action
        self.updatePlayerPos(update_speed, rotation, dt, ignore_step=ignore_step)

        return task.cont
