#!/usr/bin/python
# -*- coding: utf-8 -*-

#
# PYTHON IMPORTS
#
import math

#
# PANDA3D ENGINE IMPORTS
#
from panda3d.core import (
    CollisionTraverser,
    CollisionHandlerEvent,
    CollisionHandlerQueue,
    CollisionHandlerPusher,
    CollisionNode,
    CollisionSphere,
    CollisionSegment,
    CollisionRay,
    Point3,
    Vec3,
    NodePath,
    BitMask32,
    ModelRoot,
    )
from panda3d.physics import (
    PhysicsCollisionHandler,
    ActorNode)

__author__ = "Fireclaw the Fox"
__license__ = """
Simplified BSD (BSD 2-Clause) License.
See License.txt or http://opensource.org/licenses/BSD-2-Clause for more info
"""


#
# PHYSICS FUNCTIONS
#
class Physics:
    def __init__(self):
        self.rayCTrav = CollisionTraverser("collision traverser for ray tests")
        self.physics_pusher = PhysicsCollisionHandler()

        self.collisionevent_handler = CollisionHandlerEvent()
        self.collisionevent_handler.addInPattern('%fn-in-%in')
        self.collisionevent_handler.addOutPattern('%fn-out-%in')
        self.collisionevent_handler.addInPattern('%fn-in')
        self.collisionevent_handler.addOutPattern('%fn-out')

        self.event_mask = BitMask32(0x80)  #1000 0000
        self.body_mask = BitMask32(0x70)  #0111 0000
        self.ray_mask = BitMask32(0x0f)  #0000 1111

        self.ignore_step = False
        self.anRemoved = False

    def startPhysics(self):
        """Start and set up the remaining physics parts of the character
        Should be called at the character setup and base start method"""
        self.actorNode = ActorNode("playerPhysicsController")
        base.physicsMgr.attachPhysicalNode(self.actorNode)
        self.actorNode.getPhysicsObject().setMass(self.player_mass)
        self.mainNode = render.attachNewNode(self.actorNode)
        self.reparentTo(self.mainNode)

        # main character cylinder/spheres
        charCollisions = self.mainNode.attachNewNode(CollisionNode("charBody"))
        charCollisions.node().addSolid(CollisionSphere(0, 0, self.player_height/4.0, self.player_height/4.0))
        charCollisions.node().addSolid(CollisionSphere(0, 0, self.player_height/4.0*3.05, self.player_height/4.0))
        charCollisions.node().setIntoCollideMask(self.body_mask)
        charCollisions.node().setFromCollideMask(self.body_mask)
        if self.show_collisions:
            charCollisions.show()
        self.physics_pusher.addCollider(charCollisions, self.mainNode)
        base.cTrav.addCollider(charCollisions, self.physics_pusher)

        eventCollider = self.mainNode.attachNewNode(CollisionNode(self.char_collision_name))
        eventCollider.node().addSolid(CollisionSphere(0, 0, self.player_height/2.0, self.player_height/2.0))
        eventCollider.node().setIntoCollideMask(self.event_mask)
        eventCollider.node().setFromCollideMask(self.event_mask)
        if self.show_collisions:
            eventCollider.show()
        base.cTrav.addCollider(eventCollider, self.collisionevent_handler)

        self.accept("{}-in".format(self.char_collision_name), self.checkCharCollisions)

        # step check ray
        self.raytest_segment = CollisionSegment(0, 1)
        self.raytest_np = render.attachNewNode(CollisionNode("testRay"))
        self.raytest_np.node().addSolid(self.raytest_segment)
        self.raytest_np.node().setIntoCollideMask(BitMask32.allOff())
        self.raytest_np.node().setFromCollideMask(self.ray_mask)
        if self.show_collisions:
            self.raytest_np.show()

        self.raytest_queue = CollisionHandlerQueue()
        self.rayCTrav.addCollider(self.raytest_np, self.raytest_queue)

        if self.use_simple_shadow:
            # shadow ray
            ray = CollisionRay(
                0, 0, 0.5,
                0, 0, -1)
            self.shadowRay = self.mainNode.attachNewNode(CollisionNode("CharShadowRay"))
            self.shadowRay.node().addSolid(ray)
            self.shadowRay.node().setIntoCollideMask(BitMask32.allOff())
            self.shadowRay.node().setFromCollideMask(self.ray_mask)
            self.shadowRayQueue = CollisionHandlerQueue()
            base.cTrav.addCollider(self.shadowRay, self.shadowRayQueue)

    def stopPhysics(self):
        """Stops the characters physics elements. Should be called at
        character cleanup"""
        self.raytest_segment.removeNode()
        self.physics_pusher.clearColliders()
        self.rayCTrav.clearColliders()
        del self.rayCTrav

    def updatePlayerPos(self, speed, heading, dt, rotation=None, rotate_around_node=None, ignore_step=False):
        """This function should be called to set the players new
        position and heading.
        speed determines the new position of the character.
        heading sets the new direction the player will face as seen from
        the camera and can be None.
        This function will process the stepping and dependend on that
        requests fall and landing states"""
        if heading is not None:
            self.mainNode.setH(camera, heading)
            self.mainNode.setP(0)
            self.mainNode.setR(0)
        self.mainNode.setFluidPos(self.mainNode, speed)
        self.ignore_step = ignore_step
        if not ignore_step:
            if self.doStep(self.state == self.STATE_IDLE):
                if self.state == self.STATE_JUMP or self.state == self.STATE_FALL:
                    self.actorNode.getPhysicsObject().setVelocity(0, 0, 0)
                    self.request(self.STATE_LAND)
            elif self.state != self.STATE_JUMP and self.state != self.STATE_FALL:
                self.request(self.STATE_FALL)

        if self.use_simple_shadow:
            self.updateCharSimpleShadow()

    def updatePlayerPosFloating(self, speed):
        """This method will update the position of the player respecting
        the global directions rather then the players local direction,
        which is useful for example if the player gets moved by a
        platform he stands on or wind or whatever external force may
        move the player around."""
        newPos = self.mainNode.getPos() + speed
        self.mainNode.setFluidPos(newPos)

    def updatePlayerPosFix(self, position, relativeTo=None):
        """This method will place the character at the given position."""
        if relativeTo is not None:
            self.mainNode.setPos(relativeTo, position)
        else:
            self.mainNode.setPos(position)

    def updatePlayerPosFloatingFlyign(self, speed):
        """This method will update the position of the player respecting
        the global directions rather then the players local direction,
        which is useful for example if the player gets moved by a
        platform he stands on or wind or whatever external force may
        move the player around.
        Note: this function will use the not physic related position
        update function and should only be used on flying modes if
        physics are disabled"""
        newPos = self.mainNode.getPos() + speed
        self.mainNode.setPos(newPos)

    def updatePlayerHpr(self, hpr):
        self.mainNode.setHpr(hpr)

    def __getHprFloatingNewPos(self, rotation, parent):
        """This function calculates the new position the character will
        get when he will be rotated around the given parent node"""
        center = parent.getPos(render)

        posvec = parent.getPos(render) - self.mainNode.getPos(render)
        posvec.setZ(0)
        circle_radius = posvec.length()

        # calculate the current angle of the player to the platform
        xdiff = self.mainNode.getX() - parent.getX()
        ydiff = self.mainNode.getY() - parent.getY()
        cur_angle_rad = math.atan2(ydiff, xdiff)
        # calculate the amount of radians that we will move the player
        rotation_rad = math.radians(rotation)
        # the new angle to which we will move the player to
        new_angle_rad = cur_angle_rad + rotation_rad

        # calculate the new position in the circle asuming the parent as
        # the center and using the previously calculated circle radius
        # and angle.
        x = parent.getX() + circle_radius * math.cos(new_angle_rad)
        y = parent.getY() + circle_radius * math.sin(new_angle_rad)

        # finally set the new position of the player
        new_pos = self.mainNode.getPos()
        new_pos.setX(x)
        new_pos.setY(y)

        return new_pos

    def updatePlayerHprFloating(self, rotation, parent):
        """This method will update the player position according to the
        rotation around the given parent node.
        NOTE: this will currently only work with rotations around the
        Z-Axis"""

        self.mainNode.setFluidPos(self.__getHprFloatingNewPos(rotation, parent))
        self.mainNode.setH(self.mainNode.getH() + rotation)

    def updatePlayerHprFloatingFlying(self, rotation, parent):
        """This method will update the player position according to the
        rotation around the given parent node.
        NOTE: this will currently only work with rotations around the
        Z-Axis"""

        self.mainNode.setPos(self.__getHprFloatingNewPos(rotation, parent))
        self.mainNode.setH(self.mainNode.getH() + rotation)

    def checkCharCollisions(self, args):
        pos = args.getSurfacePoint(self.mainNode)

        if self.ignore_step:
            pass
        elif self.doStep(self.state == self.STATE_IDLE):
            if self.state == self.STATE_JUMP or self.state == self.STATE_FALL:
                self.actorNode.getPhysicsObject().setVelocity(0, 0, 0)
                self.request(self.STATE_LAND)
                return
        elif self.state != self.STATE_JUMP and self.state != self.STATE_FALL:
            self.request(self.STATE_FALL)
            return

    def __findPlatformRoot(self, platform):
        if platform.hasParent():
            if platform.node().getType() == ModelRoot:
                return platform
            return self.__findPlatformRoot(platform.getParent())
        return platform

    def doStep(self, preventSlipping=False):
        """This method will process the characters downward stepping to
        prevent it from floating. It will also check if the character
        landed on a movable platform and set it as active platform."""
        if self.state not in self.ignore_step_states:
            # do the step height check
            tmp_np = self.mainNode.attachNewNode("temporary")
            tmp_np.setPos(self.mainNode, 0, 0, -self.stepheight)
            point_a = self.mainNode.getPos(render)
            point_a.setZ(point_a.getZ() + self.player_height/1.8)
            point_b = tmp_np.getPos(render)
            tmp_np.removeNode()
            if point_a == point_b: return
            char_step_collision = self.getFirstCollisionEntryInLine(point_a, point_b)

            # Check if we land on a movable platform
            groundNode = self.getFirstCollisionIntoNodeInLine(point_a, point_b)
            if groundNode is not None:
                if groundNode.getName().startswith(self.platform_collision_prefix):
                    p = render.find("**/%s"%groundNode.getName())
                    self.active_platform = self.__findPlatformRoot(p)
                else:
                    self.active_platform = None
            else:
                self.active_platform = None

            # prevent slipping
            if preventSlipping and char_step_collision is not None:
                floor_normal = char_step_collision.getSurfaceNormal(render)
                zx = math.atan2(floor_normal.getZ(), floor_normal.getX())*180/math.pi
                zy = math.atan2(floor_normal.getZ(), floor_normal.getY())*180/math.pi
                zx = abs(zx-90)
                zy = abs(zy-90)
                if zy <= self.slip_free_angle and zx <= self.slip_free_angle:
                    # prevent slipping
                    if zy > 0 or zx > 0:
                        self.toggleFlyMode(True)
                    return True
            self.toggleFlyMode(False)

            if char_step_collision is not None:
                if char_step_collision.hasSurfacePoint():
                    pos = char_step_collision.getSurfacePoint(render)
                    self.mainNode.setFluidZ(pos.getZ())
                    return True
            return False
        elif self.anRemoved and self.state not in self.flying_states:
            self.anRemoved = False
            base.physicsMgr.attachPhysicalNode(self.actorNode)
        elif self.state in self.ledge_grab_states:
            tmp_np = self.mainNode.attachNewNode("temporary")
            tmp_np.setPos(self.mainNode, 0, -self.ledge_forward_check_dist, self.ledge_top_check_dist)
            ledge_point_a = tmp_np.getPos(render)
            ledge_point_b = tmp_np.getPos(render)
            ledge_point_b.setZ(self.mainNode.getZ(render) + self.ledge_bottom_check_dist)
            ledge_collision = self.getFirstCollisionEntryInLine(ledge_point_a, ledge_point_b)
            tmp_np.removeNode()

            # Check if we land on a movable platform
            ledge_node = self.getFirstCollisionIntoNodeInLine(ledge_point_a, ledge_point_b)
            if ledge_node is not None:
                if ledge_node.getName().startswith(self.platform_collision_prefix):
                    p = render.find("**/%s"%ledge_node.getName())
                    self.active_platform = self.__findPlatformRoot(p)
                else:
                    self.active_platform = None
            else:
                self.active_platform = None
            return True
        self.active_platform = None
        return False

    def toggleFlyMode(self, flyActive):
        if flyActive:
            if not self.anRemoved:
                self.anRemoved = True
                self.actorNode.getPhysicsObject().setVelocity(0,0,0)
                base.physicsMgr.removePhysicalNode(self.actorNode)
        else:
            if self.anRemoved:
                self.anRemoved = False
                base.physicsMgr.attachPhysicalNode(self.actorNode)

    def getFirstCollisionEntryInLine(self, pointA, pointB):
        """A simple raycast check which will return the collision entry
        of the first collision point as seen from point A towards point B"""
        self.raytest_segment.setPointA(pointA)
        self.raytest_segment.setPointB(pointB)
        self.rayCTrav.traverse(render)
        if self.raytest_queue.getNumEntries() <= 0:
            return None
        self.raytest_queue.sortEntries()
        entry = None
        entry = self.raytest_queue.getEntry(0)
        return entry

    def getFirstCollisionIntoNodeInLine(self, pointA, pointB):
        """A simple raycast check which will return the into node of the
        first collision point as seen from point A towards pointB"""
        entry = self.getFirstCollisionEntryInLine(pointA, pointB)
        if entry is None: return None
        node = None
        node = entry.getIntoNode()
        return node

    def getFirstCollisionInLine(self, pointA, pointB):
        """A simple raycast check which will return the first collision
        point as seen from point A towards pointB"""
        entry = self.getFirstCollisionEntryInLine(pointA, pointB)
        if entry is None: return None
        pos = None
        if entry.hasSurfacePoint():
            pos = entry.getSurfacePoint(render)
        return pos

    def updateCharSimpleShadow(self):
        """This function will update the simple shadow image below the
        character. It will synch it's position with the player as well
        as calculate it's size when the player is further away from the
        shadow/ground"""
        self.shadowRayQueue.sortEntries()
        pos = None
        if self.shadowRayQueue.getNumEntries() > 0:
            pos = self.shadowRayQueue.getEntry(0).getSurfacePoint(render)
            #normal = self.shadowRayQueue.getEntry(0).getSurfaceNormal(render)
        if pos is not None:
            self.shadow.setPos(pos.getX(), pos.getY(), pos.getZ() + 0.05)
            z_a = pos.getZ()
            z_b = self.mainNode.getZ()
            dist = z_b - z_a
            # check if we should scale the shadow below the character
            if dist <= 0:
                self.shadow.setScale(self.max_shadow_scale)
            else:
                # calculate the shadows scale from its current distance
                # to the player
                scale = self.shadow_min_scale_dist - dist / self.shadow_scale_factor
                if scale < self.min_shadow_scale:
                    scale = self.min_shadow_scale
                elif scale > self.max_shadow_scale:
                    scale = self.max_shadow_scale
                self.shadow.setScale(scale)

    def doJump(self, forwardSpeed, extraSpeedVec=Vec3()):
        """This will let the actor node jump forward on the local y-axis
        with the upward speed given in jumpForce and forward given in speed.
        Note, if the actorNode shouldn't slide after landing call the
        physics.land function with the same actorNode"""
        # as we leave the ground set the active platform, if any, to None
        self.active_platform = None
        jumpVec = Vec3(0, -forwardSpeed/self.current_max_accleration*self.jump_forward_force_mult, 1)
        jumpVec *= self.jump_strength

        # rotate the extraSpeedVector to face the same direction the mainNode vector
        charVec = self.mainNode.getRelativeVector(render, extraSpeedVec)
        charVec.normalize()
        rotatedExtraSpeedVec = charVec * extraSpeedVec.length()

        jumpVec += rotatedExtraSpeedVec
        self.actorNode.getPhysicsObject().addLocalImpulse(jumpVec)
