#!/usr/bin/python
# -*- coding: utf-8 -*-

#
# PYTHON IMPORTS
#
import logging

#
# PANDA3D ENGINE IMPORTS
#
from direct.actor.Actor import Actor
from direct.fsm.FSM import FSM
from direct.showbase.DirectObject import DirectObject
from panda3d.core import WindowProperties
from direct.gui.OnscreenImage import OnscreenImage

#
# CHARACTER SPECIFIC IMPORTS
#
from Config import Config, USEBULLET, USEINTERNAL
from CameraThirdPerson import CameraThirdPerson
from CameraFirstPerson import CameraFirstPerson
from Control import Control
if USEBULLET:
    from PhysicsBullet import Physics
elif USEINTERNAL:
    from PhysicsInternal import Physics
from Animator import Animator

__author__ = "Fireclaw the Fox"
__license__ = """
Simplified BSD (BSD 2-Clause) License.
See License.txt or http://opensource.org/licenses/BSD-2-Clause for more info
"""

class PlayerController(FSM, Config, Physics, Actor, Control, Animator):
    """This is the main class for the extended character controller.
    From here you can control, start/stop and pause the character.

    To use it, call it like the following:

    player = PlayerController(self.world)
    player.startPlayer()
    player.setStartPos(Point3f(0, 0, 2))

    You can use the Config.py script to change the behavior and set
    configuration variables to fit your needs.
    """

    # Names of the animations of the actor class
    IDLE = "Idle"
    WALK = "Walk"
    RUN = "Run"
    SPRINT = "Sprint"
    JUMP_START = "JumpStart"
    JUMP_LAND = "JumpLand"
    FALL = "Fall"
    CROUCH = "Crouch"
    CROUCH_IDLE = "Crouch_Idle"
    CRAWL = "Crawl"
    CRAWL_IDLE = "Crawl_Idle"
    WALLRUN_LEFT = "WR_Left"
    WALLRUN_RIGHT = "WR_Right"
    WALLRUN_UP = "WR_Up"
    ROLL = "Roll"
    LEDGE_GRAB = "Ledge_Grab"
    LEDGE_GRAB_UP = "LG_Up"
    LEDGE_GRAB_LEFT = "LG_Left"
    LEDGE_GRAB_RIGHT = "LG_Right"

    # PLAYER CLASS FSM STATES
    STATE_IDLE = "Idle"
    STATE_IDLE_TO_WALK = "IdleToWalk"
    STATE_IDLE_TO_RUN = "IdleToRun"
    STATE_IDLE_TO_SPRINT = "IdleToSprint"
    STATE_WALK = "Walk"
    STATE_WALK_TO_IDLE = "WalkToIdle"
    STATE_WALK_TO_RUN = "WalkToRun"
    STATE_RUN = "Run"
    STATE_RUN_TO_IDLE = "RunToIdle"
    STATE_RUN_TO_WALK = "RunToWalk"
    STATE_RUN_TO_SPRINT = "RunToSprint"
    STATE_SPRINT = "Sprint"
    STATE_SPRINT_TO_IDLE = "SprintToIdle"
    STATE_SPRINT_TO_RUN = "SprintToRun"
    STATE_JUMP = "Jump"
    STATE_LAND = "Land"
    STATE_FALL = "Fall"
    STATE_WALL_RUN = "WallRun"
    STATE_RUN_TO_WALL_RUN = "RunToWallRun"
    STATE_SPRINT_TO_WALL_RUN = "SprintToWallRun"
    STATE_LEDGE_GRAB = "LedgeGrab"
    STATE_LEDGE_GRAB_UP = "LedgeGrabUp"
    STATE_LEDGE_GRAB_LEFT = "LedgeGrabLeft"
    STATE_LEDGE_GRAB_RIGHT = "LedgeGrabRight"

    INTEL_ACTION_WALL_RUN = "WallRun"
    INTEL_ACTION_LEDGE_GRAB = "LedgeGrab"

    def __init__(self, physic_world, config=None):
        logging.info("INIT PLAYER...")
        logging.info("INIT CONFIG...")
        Config.__init__(self)
        # additional initial configuration settings set by the outher application
        self.physic_world = physic_world
        logging.info("INIT PHYSICS...")
        Physics.__init__(self)
        logging.info("INIT CONTROLS...")
        Control.__init__(self)
        logging.info("INIT CAMERA HANDLER...")
        if self.first_pserson_mode:
            logging.info("INIT FIRST PERSON CAMERA...")
            self.camera_handler = CameraFirstPerson(
                self,
                self.cam_near_clip,
                self.cam_far_clip,
                self.cam_fov)
        else:
            logging.info("INIT THIRD PERSON CAMERA...")
            self.camera_handler = CameraThirdPerson(
                self,
                self.cam_near_clip,
                self.cam_far_clip,
                self.cam_fov)
        logging.info("INIT ANIMATOR...")
        Animator.__init__(self)
        logging.info("INIT PLAYER DONE")

        #
        # STATES SETUP
        #
        self.on_ground_states = [
            self.STATE_IDLE,
            self.STATE_IDLE_TO_WALK,
            self.STATE_IDLE_TO_RUN,
            self.STATE_IDLE_TO_SPRINT,
            self.STATE_WALK,
            self.STATE_WALK_TO_IDLE,
            self.STATE_WALK_TO_RUN,
            self.STATE_RUN,
            self.STATE_RUN_TO_IDLE,
            self.STATE_RUN_TO_WALK,
            self.STATE_RUN_TO_SPRINT,
            self.STATE_SPRINT,
            self.STATE_SPRINT_TO_IDLE,
            self.STATE_SPRINT_TO_RUN,
            self.STATE_LAND,
            self.STATE_WALL_RUN,
            self.STATE_RUN_TO_WALL_RUN,
            self.STATE_SPRINT_TO_WALL_RUN,
            ]
        self.wall_run_states = [
            self.STATE_WALL_RUN,
            self.STATE_RUN_TO_WALL_RUN,
            self.STATE_SPRINT_TO_WALL_RUN,
            ]
        self.ignore_input_states = [
            self.STATE_LEDGE_GRAB_UP]
        self.ledge_grab_states = [
            self.STATE_LEDGE_GRAB,
            self.STATE_LEDGE_GRAB_LEFT,
            self.STATE_LEDGE_GRAB_RIGHT]
        self.jump_and_fall_states = [
            self.STATE_JUMP,
            self.STATE_FALL]
        # from these states a wall run can be initiated
        self.wall_run_possible_states = [
            self.STATE_RUN,
            self.STATE_RUN_TO_SPRINT,
            self.STATE_SPRINT,
            self.STATE_SPRINT_TO_RUN,
            ] + self.wall_run_states + self.jump_and_fall_states
        self.ledge_grab_possible_states = self.wall_run_states + self.jump_and_fall_states
        self.flying_states = self.ledge_grab_states
        self.ignore_step_states = [
            self.STATE_JUMP,
            ] + self.flying_states
        # set the possible transition in the FSM
        self.defaultTransitions = {
            self.STATE_IDLE: [
                self.STATE_IDLE_TO_WALK,
                self.STATE_IDLE_TO_RUN,
                self.STATE_IDLE_TO_SPRINT,
                self.STATE_LAND]
                + self.jump_and_fall_states,
            self.STATE_IDLE_TO_WALK: [
                self.STATE_IDLE,
                self.STATE_WALK]
                + self.jump_and_fall_states,
            self.STATE_IDLE_TO_RUN: [
                self.STATE_IDLE,
                self.STATE_RUN,
                self.STATE_JUMP,
                self.STATE_FALL],
            self.STATE_IDLE_TO_SPRINT: [
                self.STATE_IDLE,
                self.STATE_SPRINT,
                self.STATE_JUMP,
                self.STATE_FALL],
            self.STATE_WALK: [
                self.STATE_IDLE,
                self.STATE_WALK_TO_IDLE,
                self.STATE_WALK_TO_RUN]
                + self.jump_and_fall_states,
            self.STATE_WALK_TO_IDLE: [
                self.STATE_IDLE,
                self.STATE_RUN,
                self.STATE_WALK]
                + self.jump_and_fall_states,
            self.STATE_WALK_TO_RUN: [
                self.STATE_IDLE,
                self.STATE_RUN]
                + self.jump_and_fall_states,
            self.STATE_RUN: [
                self.STATE_IDLE,
                self.STATE_RUN_TO_IDLE,
                self.STATE_RUN_TO_WALK,
                self.STATE_RUN_TO_SPRINT]
                + self.jump_and_fall_states
                + self.wall_run_states,
            self.STATE_RUN_TO_WALK: [
                self.STATE_IDLE,
                self.STATE_WALK]
                + self.jump_and_fall_states,
            self.STATE_RUN_TO_IDLE: [
                self.STATE_IDLE,
                self.STATE_WALK,
                self.STATE_RUN,
                self.STATE_RUN_TO_SPRINT]
                + self.jump_and_fall_states,
            self.STATE_RUN_TO_SPRINT: [
                self.STATE_IDLE,
                self.STATE_RUN,
                self.STATE_SPRINT]
                + self.jump_and_fall_states
                + self.wall_run_states,
            self.STATE_SPRINT: [
                self.STATE_IDLE,
                self.STATE_SPRINT_TO_IDLE,
                self.STATE_SPRINT_TO_RUN]
                + self.jump_and_fall_states
                + self.wall_run_states,
            self.STATE_SPRINT_TO_IDLE: [
                self.STATE_IDLE,
                self.STATE_WALK,
                self.STATE_RUN,
                self.STATE_SPRINT,
                self.STATE_SPRINT_TO_RUN]
                + self.jump_and_fall_states,
            self.STATE_SPRINT_TO_RUN: [
                self.STATE_IDLE,
                self.STATE_WALK,
                self.STATE_RUN,
                self.STATE_SPRINT]
                + self.jump_and_fall_states
                + self.wall_run_states,
            self.STATE_JUMP: [
                self.STATE_FALL,
                self.STATE_LAND]
                + self.ledge_grab_states
                + self.wall_run_states,
            self.STATE_LAND: [
                self.STATE_IDLE,
                self.STATE_WALK,
                self.STATE_RUN,
                self.STATE_SPRINT,
                self.STATE_FALL],
            self.STATE_FALL: [
                self.STATE_LAND,
                self.STATE_JUMP]
                + self.ledge_grab_states
                + self.wall_run_states,
            self.STATE_WALL_RUN: [
                self.STATE_IDLE,
                self.STATE_WALK,
                self.STATE_RUN,
                self.STATE_SPRINT,
                self.STATE_LAND]
                + self.ledge_grab_states
                + self.jump_and_fall_states
                + self.wall_run_states,
            self.STATE_RUN_TO_WALL_RUN: [
                self.STATE_IDLE,
                self.STATE_RUN,
                self.STATE_SPRINT]
                + self.ledge_grab_states
                + self.jump_and_fall_states
                + self.wall_run_states,
            self.STATE_SPRINT_TO_WALL_RUN: [
                self.STATE_IDLE,
                self.STATE_RUN,
                self.STATE_SPRINT]
                + self.ledge_grab_states
                + self.jump_and_fall_states
                + self.wall_run_states,
            self.STATE_LEDGE_GRAB: [
                self.STATE_FALL,
                self.STATE_LEDGE_GRAB_UP,
                self.STATE_LEDGE_GRAB_LEFT,
                self.STATE_LEDGE_GRAB_RIGHT],
            self.STATE_LEDGE_GRAB_UP: [
                self.STATE_IDLE],
            self.STATE_LEDGE_GRAB_LEFT: [
                self.STATE_LEDGE_GRAB_UP]
                + self.ledge_grab_states
                + self.jump_and_fall_states,
            self.STATE_LEDGE_GRAB_RIGHT: [
                self.STATE_LEDGE_GRAB_UP]
                + self.ledge_grab_states
                + self.jump_and_fall_states
            }

        #
        # ACTOR SETUP
        #
        Actor.__init__(
            self,
            self.model,
            {
                self.IDLE: self.anim_idle,
                self.WALK: self.anim_walk,
                self.RUN: self.anim_run,
                self.SPRINT: self.anim_sprint,
                self.JUMP_START: self.anim_jumpstart,
                self.JUMP_LAND: self.anim_jumpland,
                self.FALL: self.anim_falling,
                self.CROUCH: self.anim_crouch_move,
                self.CROUCH_IDLE: self.anim_crouch_idle,
                self.CRAWL: self.anim_crawl_move,
                self.CRAWL_IDLE: self.anim_crawl_idle,
                self.WALLRUN_LEFT: self.anim_wallrun_left,
                self.WALLRUN_RIGHT: self.anim_wallrun_right,
                self.WALLRUN_UP: self.anim_wallrun_up,
                self.ROLL: self.anim_roll,
                self.LEDGE_GRAB: self.anim_ledge_grab,
                self.LEDGE_GRAB_UP: self.anim_ledge_grab_up,
                self.LEDGE_GRAB_LEFT: self.anim_ledge_grab_left,
                self.LEDGE_GRAB_RIGHT: self.anim_ledge_grab_right,
            })
        self.setBlend(frameBlend=self.enable_interpolation)
        logging.info("INIT FSM...")
        FSM.__init__(self, "FSM-Player")

        #
        # SHADOW SETUP
        #
        if self.use_simple_shadow:
            # setup simple shadow
            logging.info("Setup simple shadow...")
            self.shadow = OnscreenImage(self.simple_shadow_image)
            self.shadow.setP(-90)
            self.shadow.setScale(self.max_shadow_scale)
            self.shadow.setPos(0,0,0.01)
            self.shadow.setTransparency(True)
            self.shadow.setBin("fixed", 10)
            self.shadow.reparentTo(render)

        #
        # CONTROLS SETUP
        #
        self.isDown = base.mouseWatcherNode.isButtonDown
        self.mainNode = self

    def find(self, searchString):
        return Actor.find(self, searchString)

    def catchCursor(self):
        """This method will center the mouse cursor on the window and
        ensures it doesn't move away from there"""
        # center the mouse in the middle of the window
        base.win.movePointer(0, self.win_width_half, self.win_height_half)
        # Set mouse mode to relative which should work best for our purpose
        wp = WindowProperties()
        # As the M_relative mouse mode breaks the mouse handling code,
        # we simply keep the absolute mode for that.
        if base.pipe.getInterfaceName() != "TinyPanda":
            wp.setMouseMode(WindowProperties.M_relative)
        else:
            wp.setMouseMode(WindowProperties.M_absolute)
        base.win.requestProperties(wp)

    def freeCursor(self):
        """This function will activate the normal mouse mode and should
        be called after a catchCursor call to release the mouse from
        the window"""
        # free the cursor
        wp = WindowProperties()
        wp.setMouseMode(WindowProperties.M_absolute)
        base.win.requestProperties(wp)

    def changeCameraSystem(self, mode):
        """This method can be used to change between the two camera
        systems, first person and third person.
        This function will stop the current and start the new camera
        system.

        mode may be either of
        firstperson
        thirdperson"""
        self.camera_handler.centerCamera()
        self.camera_handler.stopCamera()
        if mode == "firstperson":
            self.first_pserson_mode = True
            logging.info("INIT FIRST PERSON CAMERA...")
            self.cam_near_clip = self.cam_near_clip_default_firstperson
            self.cam_fov = self.cam_fov_default_firstperson
            self.mouse_speed_x = self.mouse_speed_x_default_firsrperson
            self.mouse_speed_y = self.mouse_speed_y_default_firsrperson
            self.camera_handler = CameraFirstPerson(
                self,
                self.cam_near_clip,
                self.cam_far_clip,
                self.cam_fov)
        elif mode == "thirdperson":
            self.first_pserson_mode = False
            logging.info("INIT THIRD PERSON CAMERA...")
            self.cam_near_clip = self.cam_near_clip_default_thirdperson
            self.cam_fov = self.cam_fov_default_thirdperson
            self.mouse_speed_x = self.mouse_speed_x_default_thirdperson
            self.mouse_speed_y = self.mouse_speed_y_default_thirdperson
            self.camera_handler = CameraThirdPerson(
                self,
                self.cam_near_clip,
                self.cam_far_clip,
                self.cam_fov)
        self.camera_handler.startCamera()
        self.camera_handler.centerCamera()

    def startPlayer(self):
        """This function must be called after a player has been set up
        to make it do anything and actually show up in the game"""
        logging.debug("start player...")
        self.show()
        self.request(self.STATE_IDLE)
        logging.debug("...start physics...")
        self.startPhysics()
        logging.debug("...start coltrol...")
        self.startControl()
        logging.debug("...start camera...")
        self.camera_handler.startCamera()
        self.camera_handler.centerCamera()
        logging.debug("...player started")
        self.catchCursor()

    def stopPlayer(self):
        """This function should be called as soon as the character isn't
        needed anymore and is ready to be completely cleaned up. This
        should not be called when you want to disable the players
        movements, use pausePlayer therefor"""
        logging.debug("stop player...")
        logging.debug("...stop control...")
        self.stopControl()
        logging.debug("...stop camera...")
        self.camera_handler.stopCamera()
        logging.debug("...stop base...")
        self.freeCursor()
        # remove the simple shadow
        self.shadow.removeNode()
        # remove the actor
        Actor.cleanup(self)
        self.removeNode()
        logging.debug("...player stop")

    def pausePlayer(self):
        """This function will stop the player from beeing controlable by
        the user but will not hide or remove the player from the game"""
        self.pauseControl()
        self.pauseAnimator()
        self.camera_handler.pauseCamera()
        self.freeCursor()

    def resumePlayer(self):
        """Resumes the player from a previous called pausePlayer call"""
        self.catchCursor()
        self.startControl()
        self.resumeAnimator()
        self.camera_handler.resumeCamera()

    def setStartPos(self, startPos):
        """Set the position of the character, should at least be called
        once to place the character in an initial position, but can also
        be used to place the character at later times."""
        self.mainNode.setPos(startPos)

    def setStartHpr(self, startHpr):
        """Set the orientation of the character, should at least be
        called once to set the characters initial orientation, but can
        also be used to set the characters orientation at later times."""
        self.mainNode.setHpr(startHpr)

