#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
from direct.showbase.ShowBase import ShowBase
from panda3d.core import (
    loadPrcFileData,
    Vec3,
    Point3,
    DirectionalLight,
    VBase4,
    AmbientLight,
    CollisionTraverser,
    CollisionPlane,
    CollisionBox,
    CollisionNode,
    Plane,
    BitMask32,)
from panda3d.bullet import (
    BulletWorld,
    BulletDebugNode,
    BulletPlaneShape,
    BulletRigidBodyNode)
from panda3d.physics import ForceNode, LinearVectorForce
from direct.interval.IntervalGlobal import Sequence, Wait
from PlayerController import PlayerController
from Config import USEBULLET, USEINTERNAL

__author__ = "Fireclaw the Fox"
__license__ = """
Simplified BSD (BSD 2-Clause) License.
See License.txt or http://opensource.org/licenses/BSD-2-Clause for more info
"""

# setup Logging
logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s %(levelname)s: %(message)s",
    filename="./demo.log",
    datefmt="%d-%m-%Y %H:%M:%S",
    filemode="w")


loadPrcFileData("","""
model-path $MAIN_DIR/testmodel
cursor-hidden 1
""")

class Main(ShowBase):
    def __init__(self):
        """initialise and start the Game"""
        ShowBase.__init__(self)

        self.accept("escape", exit)
        self.accept("f1", self.toggleDebug)
        self.accept("r", self.resetPlayer)
        self.accept("p", self.togglePause)
        self.accept("f2", self.toggleCamera)
        self.disableMouse()

        #
        # SIMPLE LEVEL SETUP
        #
        self.level = loader.loadModel("../../data/models/level")
        self.level.reparentTo(render)
        #
        # Lights
        #
        alight = AmbientLight("Ambient")
        alight.setColor(VBase4(0.2, 0.2, 0.2, 1))
        alnp = render.attachNewNode(alight)
        render.setLight(alnp)
        sun = DirectionalLight("Sun")
        sun.setColor(VBase4(0.8, 0.8, 0.5, 1))
        sunnp = render.attachNewNode(sun)
        sunnp.setHpr(10, -60, 0)
        render.setLight(sunnp)
        #
        # LEVEL SETUP END
        #

        #
        # SIMPLE PHYSICS SETUP
        #
        self.useBullet = USEBULLET
        self.useInternal = USEINTERNAL
        self.debugactive = True
        #
        # BULLET
        #
        if self.useBullet:
            self.world = BulletWorld()
            self.world.setGravity(Vec3(0, 0, -9.81))
            # show the debug geometry for bullet collisions
            self.debugactive = True
            debugNode = BulletDebugNode("Debug")
            debugNode.showWireframe(True)
            debugNode.showConstraints(True)
            debugNode.showBoundingBoxes(False)
            debugNode.showNormals(True)
            debugNP = render.attachNewNode(debugNode)
            debugNP.show()
            self.world.setDebugNode(debugNode)
            shape = BulletPlaneShape(Vec3(0, 0, 0), 1)
            node = BulletRigidBodyNode("Ground")
            node.addShape(shape)
            np = render.attachNewNode(node)
            np.setPos(0, 0, -2)
            self.world.attachRigidBody(node)
            self.__taskName = "task_physicsUpdater_Bullet"
            taskMgr.add(self.updatePhysicsBullet, self.__taskName, priority=-20)
        #
        # INTERNAL
        #
        if self.useInternal:
            # enable physics
            base.enableParticles()
            base.cTrav = CollisionTraverser("base collision traverser")
            base.cTrav.setRespectPrevTransform(True)
            # setup default gravity
            gravityFN = ForceNode("world-forces")
            gravityFNP = render.attachNewNode(gravityFN)
            gravityForce = LinearVectorForce(0, 0, -9.81)  # gravity acceleration
            gravityFN.addForce(gravityForce)
            base.physicsMgr.addLinearForce(gravityForce)

            # Ground Plane
            plane = CollisionPlane(Plane(Vec3(0, 0, 1), Point3(0, 0, -2)))
            self.ground = render.attachNewNode(CollisionNode("Ground"))
            self.ground.node().addSolid(plane)
            self.ground.show()

            self.platformIntervals = []
            self.platforms = []
            self.addFloatingPlatform(0, 8.0, self.level.find("**/PlatformPos.000").getPos(), self.level.find("**/PlatformPos.001").getPos())
            self.addFloatingPlatform(1, 8.0, self.level.find("**/PlatformPos.002").getPos(), self.level.find("**/PlatformPos.003").getPos())
            self.addFloatingPlatform(2, 8.0, self.level.find("**/PlatformPos.004").getPos(), self.level.find("**/PlatformPos.005").getPos())
            self.addFloatingPlatform(3, 8.0, self.level.find("**/PlatformPos.006").getPos(), self.level.find("**/PlatformPos.007").getPos())

            self.addFloatingPlatform(4, 10.0, (0, -15, 0), (0, -15, 0), 0, (360, 0, 0))

            for ival in self.platformIntervals:
                ival.loop()

            # Intangible blocks (as used for example for collectible or event spheres)
            self.moveThroughBoxes = render.attachNewNode(CollisionNode("Ghosts"))
            box = CollisionBox((0, 0, 0.5), 1, 1, 1)
            box.setTangible(False)
            self.moveThroughBoxes.node().addSolid(box)
            # should only collide with the event sphere of the character
            self.moveThroughBoxes.node().setFromCollideMask(BitMask32.allOff())
            self.moveThroughBoxes.node().setIntoCollideMask(BitMask32(0x80))  # 1000 0000
            self.moveThroughBoxes.show()

            # Set the world
            self.world = base.cTrav

        #
        # PHYSICS SETUP END
        #

        #
        # DEBUGGING
        #
        from direct.showbase.OnScreenDebug import OnScreenDebug
        self.osd = OnScreenDebug()
        self.osd.enabled = True
        self.osd.append("Debug OSD")
        self.osd.load()
        self.osd.render()
        taskMgr.add(self.debugOSDUpdater, "update OSD")

        #
        # THE CHARACTER
        #
        self.p = PlayerController(self.world)
        self.p.startPlayer()
        self.p.setStartPos(self.level.find("**/StartPos").getPos())
        self.p.setStartHpr(self.level.find("**/StartPos").getHpr())

        self.pause = False
        #self.p.pausePlayer()
        #self.p.stopPlayer()

    def toggleDebug(self):
        if not self.debugactive:
            if self.useBullet:
                # activate phyiscs debugging
                self.debugNP.show()
            if self.useInternal:
                self.moveThroughBoxes.show()
            self.debugactive = True
        else:
            if self.useBullet:
                # deactivate phyiscs debugging
                self.debugNP.hide()
            if self.useInternal:
                self.moveThroughBoxes.hide()
            self.debugactive = False

    def resetPlayer(self):
        self.p.setStartPos(self.level.find("**/StartPos").getPos())
        self.p.setStartHpr(self.level.find("**/StartPos").getHpr())
        self.p.camera_handler.centerCamera()

    def togglePause(self):
        if self.pause:
            self.p.resumePlayer()
        else:
            self.p.pausePlayer()
        self.pause = not self.pause

    def toggleCamera(self):
        if self.p.first_pserson_mode:
            self.p.changeCameraSystem("thirdperson")
        else:
            self.p.changeCameraSystem("firstperson")

    def debugOSDUpdater(self, task):
        """Update the OSD with constantly changing values"""
        #self.osd.add("Rotation", str(self.platforms[4].getH()))
        self.osd.render()
        return task.cont

    def updatePhysicsBullet(self, task):
        """This task will handle the actualisation of
        the physic calculations each frame for the
        Bullet engine"""
        dt = globalClock.getDt()
        self.world.doPhysics(dt, 10, 1.0/180.0)
        return task.cont

    def addFloatingPlatform(self, platformID, time, platformStartPos, platformEndPos, platformStartHpr=0, platformEndHpr=0):
        # load and place the platform
        floatingPlatform = loader.loadModel("../../data/models/FloatingPlatform")
        floatingPlatform.setName(floatingPlatform.getName()+str(platformID))
        floatingPlatform.setPos(platformStartPos)
        floatingPlatform.setH(platformStartHpr)
        # rename the collision object so we can determine on which platform the character landed
        fpSub = floatingPlatform.find("**/FloatingPlatform")
        fpSub.setName(fpSub.getName()+str(platformID))
        floatingPlatform.reparentTo(self.level)

        # create the platforms movement using an interval sequence
        platformIval = Sequence(
            floatingPlatform.posInterval(time, platformEndPos, name="Platform%dTo"%platformID),
            Wait(3.0),
            floatingPlatform.posInterval(time, platformStartPos, name="Platform%dFrom"%platformID),
            Wait(3.0),
            name="platform-move-interval-%d"%platformID)

        platformHprInterval = None
        if platformEndHpr != 0:
            platformHprInterval = Sequence(
                floatingPlatform.hprInterval(time, platformEndHpr, name="Platform%dRotate"%platformID),
                name="platform-hpr-interval-%d"%platformID)

        # store the platform and its interval
        self.platforms.append(floatingPlatform)
        self.platformIntervals.append(platformIval)
        if platformHprInterval is not None:
            self.platformIntervals.append(platformHprInterval)

APP = Main()
APP.run()
